#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>

/*

#     #     #     #        #    #  
#  #  #    # #    #        #   #   
#  #  #   #   #   #        #  #    
#  #  #  #     #  #        ###     
#  #  #  #######  #        #  #    
#  #  #  #     #  #        #   #   
 ## ##   #     #  #######  #    #  


by CampbellWare,
1983, 2005, 2007

MS-DOS "walk" utility. This utility walks down a directory hierarchy,
starting  at the directory indicated on the command line, and executes a
user supplied shell command at each level of the hierarchy. 

This  utility  is  patterned  after  the  iRMX86  Human Interface "walk"
command created by CampbellWare in 1983. 

Usage: walk [-h | --help | -v || --verbose ] start_dir cmd_to_execute

*/

/* .page */

/*
============================
Global definitions, typedefs
============================
*/

#define begin {
#define end }
 
#define FALSE 0
#define TRUE  !FALSE

#define MAX_PATH_LEN  4096    /* Buffer overflow limit - max size file name  */
#define MAX_CMD_SIZE  1024    /* Buffer overflow limit - max command size    */

#define MAX_DIR_ENTRIES 1024 /* Max number of entries in a dir - arbitrary  */
#define MAX_FN_LEN    16     /* Max length of a file name wo directory part */

typedef int BOOL;

struct str_dir_entry {                 /* A directory entry                 */
   char name [ MAX_FN_LEN ];           /* Its name                          */
   BOOL is_directory;                  /* Whether it is a directory         */
   long num_lines;                     /* Number of lines, if a file        */
   long num_chars;                     /* Number of chars, if a file        */
   BOOL changed;                       /* Changed, if a file                */
};

struct str_entry_list {                /* An array of directory entries     */
   struct str_dir_entry entry [MAX_DIR_ENTRIES];
};


/*
=====================
External declarations
=====================
*/

extern  void exit      ();
extern  void *malloc   ();

#include "strprim.ext"
 
/*
============================
Global variables
============================
*/

BOOL verbose;                            /* Was verbose mode requested ?     */
int  directories_visited = 0;            /* Number of directories visited    */

/* .page */

void sort_dir_listing ( dir_listing, num_entries )

/*****************************************************************************
 *                                                                           *
 * This function accepts the directory list from scandir and sorts the       *
 * entries into alphabetic order using the bubble sort technique. The sort   *
 * is done inline, with the sorted results returned in the same space as the *
 * original information.                                                     *
 *                                                                           *
 *****************************************************************************/

struct dirent **dir_listing;        /* The dir listing from scandir          */
int    num_entries;                 /* The number of entries in the list     */

begin
   register int idx;                      /* Index into dir listing          */
   register int flips;                    /* Num flips done this loop        */
   register int iterations = 0;           /* Just for safety                 */
   register struct dirent *tmp_listing;   /* Used as part of flipping        */
   
   do
   begin
   
      idx = flips = 0;
      
      do
      begin
         /* Ripple down the list one position at a time, comparing the       */
         /* current and next entries. If the current is "larger" than the    */
         /* next, flip them and carry on, until the end of the list is hit   */
         
         if ( strcmp(dir_listing[idx]->d_name, dir_listing[idx+1]->d_name) > 0 )
         begin
            /* The first of the two is lexically bigger than the second.     */
            /* Flip them and carry on.                                       */
            
            tmp_listing        = dir_listing[idx];
            dir_listing[idx]   = dir_listing[idx+1];
            dir_listing[idx+1] = tmp_listing;
            
            flips += 1;
         end
         
      end
      while ( ++idx + 1 < num_entries );
      
      /* Continue on until there are no more flips to be made, or we have   */
      /* exceeded a "reasonable" limit for iterations and are clearly just  */
      /* thrashing - 15,000 is entirely arbitrary                           */
      
   end
   while ( ( flips > 0 ) && ( ++iterations < 15000 ) );
   
end /* sort_directory_listing */

/* .page */

struct dirent **scandir ( directory, count )

/*****************************************************************************
 *                                                                           *
 * This function accepts a directory name for a directory in the current     *
 * working directory, and creates an array of directory entry segments,      *
 * it fills with the names of all files in the directory. This is the return *
 * result of this function.                                                  *
 * parameters. Either or both may be nil strings if this is appropriate.     *
 *                                                                           *
 *****************************************************************************/

char directory [];      /* The name of the directory                         */
int  *count;            /* The number of directory entries read              */

begin

   struct dirent *dir_entry;    /* Will malloc one of these per entry       */
   struct dirent *this_entry;   /* The malloc'd segment for current entry   */
   struct dirent **all_entries; /* Array of *dir_entry                      */

   DIR *dir_conn;               /* File connection to curr directory        */

   /* Initialize return values for safety */

   *count = 0;
   all_entries = NULL;

   dir_conn = opendir ( directory );

   if ( dir_conn != NULL )
   begin
      /* We were able to open the supplied directory. Allocate a sgment to  */
      /* hold the individual entries. In the DOS port of VE v3.5, we        */
      /* arbitrarily limit this to 1024 entries. Further entries will       */
      /* simply be ignored.                                                 */

      all_entries = malloc ( MAX_DIR_ENTRIES * sizeof(dir_entry) );

      while ( (dir_entry = readdir ( dir_conn )) != NULL )
      begin
         this_entry = (struct dirent *) malloc ( sizeof(struct dirent) );
         movb ( dir_entry, this_entry, sizeof(struct dirent) );
         all_entries[(*count)++] = this_entry;

         if ( *count == MAX_DIR_ENTRIES )
            break;
      end

   end /* Were able to open the supplied directory */

   sort_dir_listing ( all_entries, *count );

   return ( all_entries );

end /* scandir */

/* .page */

void parse_arguments ( argc, argv, help, verbose, start_dir, cmd_to_execute )

/*
==============================================================================

This  function  parses  the  user supplied command line arguments, and returns
the results as the variables help, verbose, start_dir and cmd_to_execute. 

The expected command syntax is:

walk [-h | --help | -v | --verbose] start_dir cmd_to_execute

Return values:
- If "help" is TRUE, the user has requested help
- If "verbose" is TRUE, the users wants walk to produce per directory messages
- If "start_dir" is NULL, this argument was lacking - walk will terminate
- If "cmd_to_execute" is NULL, this argument was lacking - walk will terminate

==============================================================================
*/

int argc;               /* Number of command line arguments present          */
char *argv[];           /* The command line arguments themselves             */
BOOL *help;             /* Was help requested                                */
BOOL *verbose;          /* Was verbose mode requested                        */
char start_dir [];      /* The directory at which to start walking           */
char cmd_to_execute []; /* The shell command to execute at each level        */

{
   register int i;                          /* Loop counter                  */
   register BOOL have_start_dir = FALSE;    /* Do we have start dir already? */

   /* Initialize the parsed results to logical "nil" */

   *help      = FALSE;
   *verbose   = FALSE;
   strcpy ( start_dir,      "\0" );
   strcpy ( cmd_to_execute, "\0" );

   /* Check for at least three tokens on command line ("walk start_dir       */
   /* cmd_to_execute"). If at least three are not supplied, present the user */
   /* with help info and exit.                                               */

   if ( argc < 3 )
   {
      *help = TRUE;
      return;
   }

   /* Iterate over each argument and parse it into its associated variable   */

   for ( i = 1; i < argc; i++ )
   {
      if ( !strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") )
         *help = TRUE;

      else if ( !strcmp(argv[i], "-v") || !strcmp(argv[i], "--verbose") )
         *verbose = TRUE;

      else
      {
         /* This is not the help or verbose switches, so it is either the    */
         /* starting directory or the command to execute. Starting directory */
         /* must always be presented first.                                  */

         if ( have_start_dir )
         {
            /* Since we have the starting directory already, this is the     */
            /* command to execute at each level.                             */

            if ( strlen(argv[i]) < MAX_CMD_SIZE )
               strcpy ( cmd_to_execute, argv[i] );
            else
            {
               printf ( "\n%s\n", "Requested command exceeds maximum length" );
               *help = TRUE;
               return;
            }
         }

         else
         {
            /* This is the starting directory */

            if ( strlen(argv[i]) < MAX_PATH_LEN )
            {
               strcpy ( start_dir, argv[i] );
               have_start_dir = TRUE;
            }
            else
            {
               printf ( "%s\n", "Start directory name exceeds maximum length" );
               *help = TRUE;
               return;
            }
         }
      }
    }
      
} /* parse_arguments */

/* .page */

int file_select ( entry )

/*
==============================================================================

This  function  examines  the current directory entry, and determines if it is
the  current  or  home directory values ("." and ".." respectively). If so, it
returns  FALSE,  else  it  returns TRUE. This is used to ensure that these two
entries  are not included as part of walk's processing, potentially leading to
an endless loop. 

==============================================================================
*/
struct dirent *entry ;

{
   if ( !strcmp(entry->d_name, ".") || !strcmp(entry->d_name, "..") )
      return ( FALSE );
   else
      return ( TRUE );

} /* file_select */

/* .page */

void walk_hierarchy ( pathname, cmd_to_execute )

/*
==============================================================================

This routine is the main portion of walk. It recursively walks down the 
directory hierarcy, executing the supplied command once per level.

==============================================================================
*/

char pathname[];
char cmd_to_execute[];

{
    char subdir_pathname [MAX_PATH_LEN];

    BOOL is_dot;            /* Is this the "." (current) directory?          */
    BOOL is_dotdot;         /* Is this the ".." (parent) directory?          */
    int i;                  /* Loop counter                                  */
    int count;              /* Number of files in current directory          */
    struct dirent **files;  /* String table of filenames in the directory    */
    int stat_res;           /* Result of status call on file                 */
    struct stat file_info;  /* File information from the stat function       */

    directories_visited += 1;

    if ( verbose )
       printf ( "%s %s \n", "Entering Directory: ", pathname );

    files = scandir ( pathname, &count );

    /* If no files found, simply return. Clearly, no further directories to  */
    /* descend into here!                                                    */

    if ( count <= 0 )
    {      
       if ( verbose ) 
       {
          printf ( "%s\n", "No files in this directory" );
          printf( "%s %s\n", "Leaving Directory: ", pathname);
       }

       return;
    }

    /* Change to the present directory, and then execute the user supplied   */
    /* command                                                               */

    if ( chdir(pathname) == 0 )
       system ( cmd_to_execute );

    /* Iterate over every entry in the directory, searching for sub          */
    /* directories. If found, we will descend into them.                     */

    for ( i = 0; i < count; i++ )
    {
       stat_res = stat ( files[i]->d_name, &file_info );

       is_dot    = ( strcmp ( files[i]->d_name, "."  ) == 0 );
       is_dotdot = ( strcmp ( files[i]->d_name, ".." ) == 0 );

       if ( S_ISDIR(file_info.st_mode) && !is_dot && !is_dotdot )
       {
          /* This is a directory, but is not the current directory or the    */
          /* the parent directory - descend into it and walk it              */

          /* Append the directory name to the pathname, and recursively call */
          /* ourselves. First check to ensure that this will not overflow    */
          /* the available filename buffer space. If so, error out.          */

           if ( strlen(pathname) + strlen(files[i]->d_name) + 1 > MAX_PATH_LEN )
           {
              /* Overflowed our maximum pathname - quit with an error */

              printf ( "\n%s\n", "Walk's maximum directory depth exceeded - error exit" );
              exit ( 1 );
           }

           /* Append the subdirectory name to the current path, and descend  */
           /* into the subdirectory.                                         */

           strcpy ( subdir_pathname, pathname );
           strcat ( subdir_pathname, "/" );
           strcat ( subdir_pathname, files[i]->d_name );

           walk_hierarchy ( subdir_pathname, cmd_to_execute );

           /* Descending into the subdirectory may have changed the current  */
           /* directory. As a safety measure, change back to the directory   */
           /* provided as the argument to this call.                         */

           chdir ( pathname );
       }

       /* scandir builds a table of malloc'd directory entries. As we finish */
       /* each entry, free the memory associated with it.                    */

       free ( files[i] );

    } /* Per file loop */

    /* The table of directory entries returned by scandir is itself a        */
    /* malloc'd segment. Free it as well now that we are all done with this  */
    /* directory.                                                            */

    free ( files );

    if ( verbose )
       printf( "%s %s\n", "Leaving Directory: ", pathname);

} /* walk_hierarchy */

/* .page */

int main ( argc, argv )   

/*
==============================================================================

This  is  the  walk  mainline.  It  initiates  parsing  of  the  command  line
arguments,  establishes  the  complete pathname of the starting directory, and
initiates  the  walk_hierarchy  process. It generates a final message when the
process is completed. 

==============================================================================
*/

int argc;
char *argv[];

{ 
    BOOL help;                              /* Has help been requested?      */
    char initial_dir    [ MAX_PATH_LEN ];   /* Directory we started in       */
    char start_dir      [ MAX_PATH_LEN ];   /* Starting directory pathname   */
    char cmd_to_execute [ MAX_CMD_SIZE ];   /* Command to execute each level */                     

    directories_visited = 0;

    parse_arguments ( argc, argv, &help, &verbose, start_dir, cmd_to_execute );

    if ( help || strlen(start_dir) == 0 || strlen(cmd_to_execute) == 0 )
    {
       /* Help has been requested - provide the same and then exit */

       printf ( "%s\n\n", "Usage: walk [-h | --help | -v | --verbose] start_dir cmd_to_execute" );
       printf ( "%s\n",   "walk descends a directory hierarchy, executing cmd_to_execute" );
       printf ( "%s\n\n", "at each level of the hierarchy. Enclose cmd_to_execute in quotes" );

       exit (0);
    }

    /* Save off the initial directory so that we can return the user there */
    /* when done.                                                          */

    getcwd ( initial_dir, MAX_PATH_LEN );

    /* Change to the supplied directory */

    if ( chdir(start_dir) )
    { 
       printf ( "%s\n", "Cannot enter start directory" );
       exit(1);
    }
    
    /* Get its full pathname */

    if ( getcwd(start_dir, MAX_PATH_LEN) == NULL )
    {
       printf ( "%s\n", "Failed to set up start dir absolute pathname" );
       exit(1);
    }

    if ( verbose )
    {
       printf ( "\n%s %s \n", "MS-DOS walk, v1.0 - Starting Directory: ", start_dir      );
       printf ( "%s %s \n\n", "Command to  be Executed Per Directory: ", cmd_to_execute );
    }

    /* OK, here goes! Lets dive into the directory hierarchy */

    walk_hierarchy ( start_dir, cmd_to_execute );

    /* Change back to initial directory, print out some final stats and we */
    /* are done                                                            */

    chdir ( initial_dir );
    printf( "\n%s %d %s\n", "walk completed.", directories_visited, "directories visited" );

} /* walk main */

